#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "utility.h"

#define USAGE "USAGE: trinuc_freqs [-asymmetry] trinuc_changes_counts\n"

#define MAX_LINE 2048

#define NUM_NUC 4     /* Size of nucleotide alphabet */

#define NUC2NUM(x) ((unsigned int)strchr(NAUC, (x)) - (unsigned int)NAUC)

typedef struct {
  int trinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  double length;
} Branch;

typedef struct {
  int asymmetry;
  int n_seq;
  int seq_length;
  int num_branches;
  int num_mutation_parameters;
  Branch* branch;
  int trinuc_parameters[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int num_mutation_set;
  int*** param_enum;
} Params;

char NAUC[7] = "TCAGN-";
char NCUC[7] = "AGTCN-";
int complement[4] = {2,3,0,1};

void read_params(int argc, char* argv[], Params* params);
void print_parameters(Params* params, double* parameters);
void read_changes(Params* params, char* changesfiles);
void make_trinuc_parameters(Params* params);
void mle(Params* params, double* parameters);

int main(int argc, char* argv[]) {
  Params *params;
  double* parameters;

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");

  read_params(argc, argv, params);

  (parameters = (double*)malloc(sizeof(double) * params->num_mutation_parameters)) || die("Could not allocate memory for parameters\n");

  mle(params, parameters);
  print_parameters(params, parameters);

  return 0;
}

void mle(Params* params, double* parameters) {
  int b, i, j;
  double numer, denom;

  for(i = 0; i < params->num_mutation_parameters; i++) {
    numer = denom = 0.0;
    for(b = 0; b < params->num_branches; b++) {
      for(j = 0; j < params->num_mutation_set; j++) {
	numer += params->branch[b].trinuc[params->param_enum[i][j][0]][params->param_enum[i][j][1]][params->param_enum[i][j][2]][params->param_enum[i][j][3]]; 
	denom += params->branch[b].length * (params->branch[b].trinuc[params->param_enum[i][j][0]][params->param_enum[i][j][1]][params->param_enum[i][j][2]][params->param_enum[i][j][3]] + params->branch[b].trinuc[params->param_enum[i][j][0]][params->param_enum[i][j][1]][params->param_enum[i][j][2]][params->param_enum[i][j][1]]);
      }
    }
    parameters[i] = numer / denom;
  }
}

void print_parameters(Params* params, double* parameters) {
  int i;

  for(i = 0; i < params->num_mutation_parameters; i++)
    if(params->asymmetry)
      printf("%c%c%c->%c = %10.15g\n",
	     NAUC[params->param_enum[i][0][0]], NAUC[params->param_enum[i][0][1]],
	     NAUC[params->param_enum[i][0][2]], NAUC[params->param_enum[i][0][3]],
	     parameters[i]);
    else
      printf("%c%c%c->%c = %c%c%c->%c = %10.15g\n",
	     NAUC[params->param_enum[i][0][0]], NAUC[params->param_enum[i][0][1]],
	     NAUC[params->param_enum[i][0][2]], NAUC[params->param_enum[i][0][3]],
	     NAUC[params->param_enum[i][1][0]], NAUC[params->param_enum[i][1][1]],
	     NAUC[params->param_enum[i][1][2]], NAUC[params->param_enum[i][1][3]],
	     parameters[i]);
}

void read_params(int argc, char* argv[], Params* params) {
  int i;

  params->asymmetry = 0;

  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    if(!strcmp(argv[i] + 1, "asymmetry"))
      params->asymmetry = 1;
    else 
      die(USAGE);
  }
  argc == i + 1 || die(USAGE);
  read_changes(params, argv[i]);
  make_trinuc_parameters(params);
}

void make_trinuc_parameters(Params* params) {
  int i, j, k, l;
  int count = 0;
  int tmp;

  if(params->asymmetry) {
    params->num_mutation_parameters = 192;
    params->num_mutation_set = 1;
  }
  else {
    params->num_mutation_parameters = 96;
    params->num_mutation_set = 2;
  }

  (params->param_enum = (int***)malloc(sizeof(int**) * params->num_mutation_parameters)) || die("Could not allocate memory for params->param_enum\n");
  for(i = 0; i < params->num_mutation_parameters; i++) {
    (params->param_enum[i] = (int**)malloc(sizeof(int*) * params->num_mutation_set)) || die("Could not allocate memory for params->param_enum[%d]\n", i);
    for(j = 0; j < params->num_mutation_set; j++)
      (params->param_enum[i][j] = (int*)malloc(sizeof(int) * 4)) || die("Could not allocate memory for params->param_enum[%d][%d]\n", i, j);
  }

  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  params->trinuc_parameters[i][j][k][l] = -1;

  tmp = params->num_mutation_set - 1;

  for(k = 0; k < NUM_NUC; k++)
    for(l = 0; l < NUM_NUC; l++)
      for(i = 0; i < NUM_NUC; i++)
	for(j = 0; j < NUM_NUC; j++)
	  if(l != k && params->trinuc_parameters[i][k][j][l] == -1) {
	    params->trinuc_parameters[i][k][j][l] = count;
	    params->param_enum[count][0][0] = i;
	    params->param_enum[count][0][1] = k;
	    params->param_enum[count][0][2] = j;
	    params->param_enum[count][0][3] = l;
	    if(params->asymmetry)
	      count++;
	    params->trinuc_parameters[complement[j]][complement[k]][complement[i]][complement[l]] = count;
	    params->param_enum[count][tmp][0] = complement[j];
	    params->param_enum[count][tmp][1] = complement[k];
	    params->param_enum[count][tmp][2] = complement[i];
	    params->param_enum[count][tmp][3] = complement[l];
	    count++;
	  }
}

void read_changes(Params* params, char* changesfile) {
  FILE* in;
  char nuc[6];
  int count;
  char line[MAX_LINE];
  int b;

  (in = fopen(changesfile, "r")) || die("Could not open changesfile: %s\n", changesfile);

  fscanf(in, "%d:", &params->num_branches) == 1 || die("Error in changesfile\n");
  (params->branch = (Branch*)malloc(sizeof(Branch) * params->num_branches)) || die("Could not allocate memory for params->branch\n");
  for(b = 0; b < params->num_branches; b++)
    fscanf(in, "%lg ", &params->branch[b].length) == 1 || die("Error in changesfile\n");
  while(fscanf(in, "%c%c%c -> %c%c%c: ", nuc, nuc+1, nuc+2, nuc+3, nuc+4, nuc+5) == 6) {
    for(b = 0; b < params->num_branches; b++)
      fscanf(in, "%d ", &params->branch[b].trinuc[NUC2NUM(nuc[0])][NUC2NUM(nuc[1])][NUC2NUM(nuc[2])][NUC2NUM(nuc[4])]) == 1 || die("Error in changesfile\n");
  }
}



